@@ -1,9 +1,10 @@
 import polars as pl
-import scipy.stats as stats
 from proxiflow.config import Config
 from proxiflow.utils import generate_trace
 from .core_utils import check_columns
 
+from typing import Dict, Any, List
+
 
 class Normalizer:
     """
@@ -17,9 +18,9 @@ class Normalizer:
         :param config: A Config object containing the normalization configuration values.
         :type config: Config
         """
-        self.config = config.normalization_config
+        self.config: Dict[str, Any] = config.normalization_config
 
-    def normalize(self, df: pl.DataFrame) -> pl.DataFrame:
+    def execute(self, df: pl.DataFrame) -> pl.DataFrame:
         """
         Normalize the specified DataFrame using the specified configuration.
 
@@ -28,11 +29,11 @@ class Normalizer:
         :return: The normalized DataFrame.
         :rtype: polars.DataFrame
         """
-        normalized_df = df.clone()
+        normalized_df = df
+
         # Apply min-max normalization
-        min_max_cols = self.config["min_max"]
+        min_max_cols: list[str] = self.config.get("min_max", [])
         if min_max_cols:
-            # Normalize the specified columns
             try:
                 normalized_df = self._min_max_normalize(normalized_df, min_max_cols)
             except Exception as e:
@@ -40,9 +41,8 @@ class Normalizer:
                 raise Exception(f"Trying min-max normalization: {trace}")
 
         # Apply z-score normalization
-        z_score_cols = self.config["z_score"]
+        z_score_cols: list[str] = self.config.get("z_score", [])
         if z_score_cols:
-            # Normalize the specified columns
             try:
                 normalized_df = self._z_score_normalize(normalized_df, z_score_cols)
             except Exception as e:
@@ -50,9 +50,8 @@ class Normalizer:
                 raise Exception(f"Trying z-score normalization: {trace}")
 
         # Apply log normalization
-        log_cols = self.config["log"]
+        log_cols: list[str] = self.config.get("log", [])
         if log_cols:
-            # Normalize the specified columns
             try:
                 normalized_df = self._log_normalize(normalized_df, log_cols)
             except Exception as e:
@@ -61,7 +60,7 @@ class Normalizer:
 
         return normalized_df
 
-    def _min_max_normalize(self, df: pl.DataFrame, columns: list[str]) -> pl.DataFrame:
+    def _min_max_normalize(self, df: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
         """
         Applies min-max normalization to the specified columns of the given DataFrame.
 
@@ -72,29 +71,35 @@ class Normalizer:
         :return: The normalized DataFrame.
         :rtype: polars.DataFrame
         """
-        clone_df = df.clone()
-        columns = check_columns(clone_df, columns)
+        columns = check_columns(df, columns)
         # If no columns exist, return the original DataFrame
         if len(columns) == 0:
-            return clone_df
-        # Select the specified columns
-        selected_df = clone_df.select(columns)
+            return df
 
-        for col in selected_df.columns:
-            # We can not subtract strings, so we only normalize numeric columns
-            if clone_df[col].dtype == pl.Int64 or clone_df[col].dtype == pl.Float64:
-                # Get the min and max values of the column
-                min_val = selected_df[col].min()
-                max_val = selected_df[col].max()
+        exprs = []
+        for col in columns:
+            # We need to check types because df[col].min()/max() are not guaranteed to be numeric types
+            dtype = df.schema[col]
+            if dtype in (pl.Int64, pl.Float64): 
+                min_val = df[col].min()
+                max_val = df[col].max()
+                if not (isinstance(min_val, (int, float)) and isinstance(max_val, (int, float))):
+                    continue  
                 if max_val - min_val == 0:
                     raise ValueError(f"Error normalizing min-max column {col}: division by zero")
-                # Normalize the column
-                min_max = (df[col] - min_val) / (max_val - min_val)
-                clone_df.replace(col, min_max)
+                exprs.append(((pl.col(col) - min_val) / (max_val - min_val)).alias(col))
+            else:
+                exprs.append(pl.col(col))
 
-        return clone_df
+        # For columns not in the normalization list, keep them unchanged
+        for col in df.columns:
+            if col not in columns:
+                exprs.append(pl.col(col))
 
-    def _z_score_normalize(self, df: pl.DataFrame, columns: list[str]) -> pl.DataFrame:
+        # Return DataFrame with normalized columns
+        return df.with_columns(exprs)
+
+    def _z_score_normalize(self, df: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
         """
         Applies z-score normalization to the specified columns of the given DataFrame.
 
@@ -105,24 +110,20 @@ class Normalizer:
         :return: The normalized DataFrame.
         :rtype: polars.DataFrame
         """
-        clone_df = df.clone()
-        columns = check_columns(clone_df, columns)
-
+        columns = check_columns(df, columns)
         if len(columns) == 0:
-            return clone_df
+            return df
 
-        selected_df = clone_df.select(columns)
+        exprs = []
+        for col in df.columns:
+            dtype = df.schema[col]
+            if col in columns and dtype in (pl.Int64, pl.Float64):
+                exprs.append(((pl.col(col) - pl.col(col).mean()) / pl.col(col).std(ddof=0)).alias(col))
+            else:
+                exprs.append(pl.col(col))
+        return df.with_columns(exprs)
 
-        for col in selected_df.columns:
-            if clone_df[col].dtype == pl.Int64 or clone_df[col].dtype == pl.Float64:
-                # Get the values of the column. Filter out None values
-                values = list(filter(lambda x: x is not None, clone_df[col].to_list()))
-                z_score = stats.zscore(values)
-                clone_df.replace(col, pl.Series(z_score))
-
-        return clone_df
-
-    def _log_normalize(self, df: pl.DataFrame, columns: list[str]) -> pl.DataFrame:
+    def _log_normalize(self, df: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
         """
         Applies log normalization to the specified columns of the given DataFrame.
 
@@ -133,18 +134,15 @@ class Normalizer:
         :return: The normalized DataFrame.
         :rtype: polars.DataFrame
         """
-        clone_df = df.clone()
-        columns = check_columns(clone_df, columns)
-
+        columns = check_columns(df, columns)
         if len(columns) == 0:
-            return clone_df
+            return df
 
-        selected_df = clone_df.select(columns)
-
-        for col in selected_df.columns:
-            if clone_df[col].dtype == pl.Int64 or clone_df[col].dtype == pl.Float64:
-                norm = (1 + clone_df[col]) / 2
-                log = norm.log()
-                clone_df.replace(col, log)
-
-        return clone_df
+        exprs = []
+        for col in df.columns:
+            dtype = df.schema[col]
+            if col in columns and dtype in (pl.Int64, pl.Float64):
+                exprs.append(((1 + pl.col(col)) / 2).log().alias(col))
+            else:
+                exprs.append(pl.col(col))
+        return df.with_columns(exprs)
